
# linkUtils.py
# Andrew Davison, ad@coe.psu.ac.th, Feb. 2025
'''
  Linkage functions for Pymunk and some 
  Pygame and Matplotlib support.
'''


import math
import pygame
from pygame.font import *

import pymunk.pygame_util
from pymunk.vec2d import Vec2d

from matplotlib import pyplot as plt


# https://www.pygame.org/docs/ref/color_list.html
BLACK = (0, 0, 0)
WHITE = (255, 255, 255)
DARKGRAY = (169, 169, 169)
LIGHTGRAY = (211, 211, 211)
RED = (255, 0, 0)
PINK = (255,202,229)
GREEN = (0,255, 0)
LIGHTGREEN = (144, 238, 144)
BLUE = (0, 0, 255)
LIGHTBLUE = (173, 216, 230)
MUSTARD = (255, 219, 88)
TANGERINE = (240, 128, 0)

FPS = 60
SHP_FILT = pymunk.ShapeFilter(group=1)  
   # for disabling collisions 

pgOptions = None
mouseJoint = None
mouse_body = pymunk.Body(body_type=pymunk.Body.KINEMATIC)
   # used for grabbing shapes with the cursor

ballsPos = []
font = None
instrs = ["k : clear curve", "<space> : pause/resume execution"]


def scrInit(title, width=800, height=700):
  # initialize pymunk to use pygame
  global pgOptions, font
  pygame.init()
  screenSize = (width, height)
  screen = pygame.display.set_mode(screenSize)
  screen.fill(WHITE)
  pygame.display.set_caption(title)
  pgOptions = pymunk.pygame_util.DrawOptions(screen)
       # use Pymunk with Pygame coordinates
  font = pygame.font.Font(None, 30)  # for laberling the balls
  return screen


def animate(screen, space, drawBall, force=5000,
            messages=instrs):
  '''
    Use space.debug_draw() to draw the Pymunk shapes.
    The line associated with drawBall is a collection
    of coordinates that are drawn with Pygame's lines()
    and the ball labels are drawn using blit().
    The user can reset those coordinates by typing 'K'
    (clear), and pause the animation by pressing <space>.
    Pymunk shapes can be 'grabbed' by using the mouse
    which generates a joint.

    At the end of the function, the drawn curve is 
    returned as a list of (x,y) tuples.
  '''
  global mouseJoint
  clock = pygame.time.Clock()
  
  pts = []
  isPaused = False
  isRunning = True
  while isRunning:  # game loop
    clock.tick(FPS)    # set Pygame loop speed
  
    # handle events
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        isRunning = False
      elif event.type == pygame.MOUSEBUTTONDOWN:
        addMouseJoint(space, event, force)
      elif event.type == pygame.MOUSEBUTTONUP:
        if mouseJoint is not None:
          space.remove(mouseJoint)
          mouseJoint = None

      if event.type == pygame.KEYDOWN:
        if event.key == pygame.K_SPACE: 
          isPaused = not isPaused   # toggle
        elif event.key == pygame.K_k: 
          pts = []   # clear pts

    # update game state
    if not isPaused:
      space.step(1/FPS)  # update physics
      mouse_body.position = pygame.mouse.get_pos()
      pts.append(drawBall.position)
   
    # redraw game
    screen.fill(WHITE)
    if len(pts) > 1:    # draw the curve
      shape = list(drawBall.shapes)[0]
      r, g, b, _ = shape.color
      pygame.draw.lines(screen, (r,g,b), False, pts)
  
    showMsgs(screen, messages)
    space.debug_draw(pgOptions)
      # draw pyMunk space using PyGame drawing options
    labelBalls(screen, ballsPos)
    pygame.display.update()
  
  pygame.quit()

  # convert Vec2d verts to (x,y) tuples
  vs = [ pt.int_tuple for pt in pts]
  return vs


def labelBalls(screen, ballsPos):
  if len(ballsPos) > 0:
    for bp in ballsPos:
      if len(bp) == 2:  # a dynamic shape with a body
        name, body = bp
        pos = (body.position + Vec2d(2,7)).int_tuple
      else:   # static; use stored coordinate
        _, name, coord = bp
        pos = (coord + Vec2d(2,7)).int_tuple
      screen.blit( font.render(name, True, BLUE), pos)


def showMsgs(screen, msgs):
  # put each message on a new line
  height = font.size(msgs[0])[1]
  scrWidth, _ = screen.get_size()
  x = scrWidth/2
  y = 10
  for msg in msgs:
    screen.blit( font.render(msg,True,LIGHTBLUE), (x,y))
    y += height


def printTuple(msg, tup):
  # for debugging
  print(f"{msg}: ({tup[0]:.2f}, {tup[1]:.2f})")


# ------- pymunk ------------------------


def addMouseJoint(space, event, force=5000):
  # use a joint to 'grab' the nearest shape to the cursor
  global mouseJoint

  if mouseJoint is not None:
    space.remove(mouseJoint)
    mouseJoint = None
  
  p = event.pos
  hit = space.point_query_nearest(p, 5, pymunk.ShapeFilter())
  if hit is not None and \
      hit.shape.body.body_type == pymunk.Body.DYNAMIC:
    shape = hit.shape
    # Use closest point on surface if click is outside shape
    if hit.distance > 0:
      nearest = hit.point
    else:
      nearest = p
    mouseJoint = pymunk.PivotJoint(
                     mouse_body, shape.body, (0, 0),
                     shape.body.world_to_local(nearest))
    mouseJoint.max_force = force
    # mouseJoint.error_bias = (1 - 0.15) ** 60
    space.add(mouseJoint)


def ball(space, name, coord, radius=10, 
                          color=DARKGRAY, solid=False):
  global ballsPos

  body = pymunk.Body()
  body.position = coord
  shape = pymunk.Circle(body, radius)
  shape.color = color + (0,)
  shape.mass = 1
  shape.elasticity = 0
  if not solid:  # default is no collision detection
    shape.filter = SHP_FILT
  space.add(body, shape)
  ballsPos.append((name, body))
  return body


def staticBall(space, name, coord):
  pt = pymunk.Body(body_type=pymunk.Body.STATIC)
  pt.position = coord
  space.add(pt)
  ballsPos.append(("static", name, coord))
  return pt



def beam(space, pt1, pt2, color=LIGHTGRAY):
  coord1 = pt1.position
  coord2 = pt2.position
  beam = segmentBody(space, coord1, coord2, color)
  pin(space, pt1, beam)
  pin(space, beam, pt2, coord2-coord1)
  return beam


def segmentBody(space, startCoord, endCoord, color):
  body = pymunk.Body()
  body.position = startCoord
  offset = endCoord - startCoord
  shape = pymunk.Segment(body, (0, 0), offset, 5)
  shape.color = color + (0,)
  shape.mass = 1
  shape.elasticity = 0
  shape.friction = 1
  shape.filter = SHP_FILT
  space.add(body, shape)
  return body


def dist(ball1, ball2):
  return ball1.position.get_distance(ball2.position)


# ------------ composite shapes ------------

def tBeam(space, pt, width, depth, name="", color=LIGHTGRAY):
  # return body and balls at three prongs of the 'T'
  # bottom, left, right
  tBody = pymunk.Body()
  center = pt.position
  tBody.position = center
  arm = pymunk.Segment(tBody, (-width,0), (width,0), 5)
  tip = pymunk.Segment(tBody, (0,0), (0,depth), 5)
  for seg in [arm, tip]:
    seg.color = color + (0,)
    seg.mass = 1
    seg.elasticity = 0
    seg.friction = 1
    seg.filter = SHP_FILT
  space.add(arm, tip, tBody)
  pin(space, tBody, pt)

  upLabel = "" if name == "" else "U"+name
  leftLabel = "" if name == "" else "L"+name
  rightLabel = "" if name == "" else "R"+name
  bot = ball(space, upLabel, center + Vec2d(0, depth))
  left = ball(space, leftLabel, center + Vec2d(-width,0))
  right = ball(space, rightLabel, center + Vec2d(width,0))
  pin(space, tBody, bot, (0,depth))
  pin(space, tBody, left, (-width, 0))
  pin(space, tBody, right, (width, 0))

  return tBody, bot, left, right


def triangle(space, pt1, pt2, pt3, color=LIGHTBLUE):
  '''
    origin in local coordinates is at pt1, and the triangle
    is connected 1-2-3, then back to 1.
  '''
  coord1 = pt1.position
  coord2 = pt2.position
  coord3 = pt3.position
  tBody = pymunk.Body()
  tBody.position = coord1
  arm1 = pymunk.Segment(tBody, (0,0), coord2-coord1, 5)
  arm2 = pymunk.Segment(tBody, coord2-coord1, coord3-coord1, 5)
  arm3 = pymunk.Segment(tBody, coord3-coord1, (0,0), 5)
  for seg in [arm1, arm2, arm3]:
    seg.color = color + (0,)
    seg.mass = 1
    seg.elasticity = 0
    seg.friction = 1
    seg.filter = SHP_FILT
  space.add(arm1, arm2, arm3, tBody)

  pin(space, tBody, pt1)
  pin(space, tBody, pt2, coord2-coord1)
  pin(space, tBody, pt3, coord3-coord1)
  return tBody


def xBeam(space, pt, width, height, color=MUSTARD):
  '''
    pt is the center of the x-beam with two width length
    arms and two height length trunks top and bottom
  '''
  xBody = pymunk.Body()
  xBody.position = pt.position
  arm = pymunk.Segment(xBody, (-width,0), (width,0), 5)
  trunk = pymunk.Segment(xBody, (0,-height), (0,height), 5)
  for seg in [arm, trunk]:
    seg.color = color + (0,)
    seg.mass = 1
    seg.elasticity = 0
    seg.friction = 1
    seg.filter = SHP_FILT
  joint = pymunk.PinJoint(xBody, pt, (0, 0), (0, 0))
  joint.collide_bodies=False
  space.add(arm, trunk, xBody, joint)
  return xBody



def lBeam(space, pt, width, height, color=MUSTARD):
  # pt is the origin with arm to right and trunk up
  lBody = pymunk.Body()
  lBody.position = pt.position
  arm = pymunk.Segment(lBody, (0,0), (width,0), 5)
  trunk = pymunk.Segment(lBody, (0,0), (0,-height), 5)
  for seg in [arm, trunk]:
    seg.color = color + (0,)
    seg.mass = 1
    seg.elasticity = 0
    seg.friction = 1
    seg.filter = SHP_FILT
  joint = pymunk.PinJoint(lBody, pt, (0, 0), (0, 0))
  joint.collide_bodies=False
  space.add(arm, trunk, lBody, joint)
  return lBody


# ----------- simplified joints -----------

def pin(space, b1, b2, pt1=(0,0)):
  joint = pymunk.PinJoint(b1, b2, pt1, (0, 0))
  joint.collide_bodies=False
  space.add(joint)


def groove(space, beam, ball, 
              startCoord=(0,0), endCoord=None):
  # default is to make groove as long as the beam's segment
  if endCoord == None:
    shape = list(beam.shapes)[0]
    if isinstance(shape, pymunk.Segment):
      endCoord = shape.b - shape.a
    else:
      print("Segment for groove not found")
  joint = pymunk.GrooveJoint(beam, ball, 
                       startCoord, endCoord, (0,0))
  joint.collide_bodies=False
  space.add(joint)
  

def rotary(space, ball, beam, minDeg, maxDeg):
  joint = pymunk.RotaryLimitJoint(ball, beam, 
            math.radians(minDeg), math.radians(maxDeg))
  joint.collide_bodies=False
  space.add(joint)


def geared(space, b1, b2, ratio=1):
  # default is to use the same gearing for both bodies
  joint = pymunk.GearJoint(b1, b2, 0, ratio)
  joint.collide_bodies=False
  space.add(joint)



def coupler(space, ball1, ball2, r, offset):
  '''
   Create two wheels of radius r centered at the
   ball positions which are geared so that they
   rotate at the same rate. Link the wheels
   with a beam whose two ends are at the offset
   coordinate relative to each wheel's center.

   Return the beam and balls representing its two ends. 
  '''
  coord1 = ball1.position
  coord2 = ball2.position
  coordC = coord1 + offset
  coordD = coord2 + offset

  ballC = ball(space, "", coordC)
  ballD = ball(space, "", coordD)

  lWheel = ball(space, "", coord1, radius=r)
  pin(space, lWheel, ball1)
  pin(space, lWheel, ballC, offset)
  
  rWheel = ball(space, "", coord2, radius=r)
  pin(space, rWheel, ball2)
  pin(space, rWheel, ballD, offset)
  
  geared(space, lWheel, rWheel)
    # geared so that the wheels will rotate at the same rate 

  beamCD = beam(space, ballC, ballD)
  return beamCD, ballC, ballD


# --------------- matplotlib ------------------

def plotPoints(pts, title):  
  xs, ys = zip(*pts)
  plt.axis('equal')
  plt.scatter(xs, ys, s=2)
  plt.title(title)
  plt.show()

