
# parabolaVS.py
# Andrew Davison, ad@coe.psu.ac.th, Feb. 2025
# Van Schooten's Parabola Drawer
# https://old.maa.org/press/periodicals/convergence/historical-activities-for-calculus-module-1-curve-drawing-then-and-now
#  - includes a nice animation
# https://www.geogebra.org/m/DDyz25cq

import math
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
from pymunk.vec2d import Vec2d
from linkUtils import *

from scipy.optimize import curve_fit


def parabolaEqu(x, a, b, c):
  return a*x**2 + b*x + c


title = "Van Schooten Parabola"
screen = scrInit(title, width=900, height=700)
scrWidth, scrHeight = screen.get_size()
space = pymunk.Space()
coord0 = Vec2d(scrWidth/2, scrHeight/2)

STEP = 60
VERT = Vec2d(0, 3*STEP)
HORIZ = Vec2d(4*STEP, 0)

coordA = Vec2d(50, 100)
coordC = Vec2d(750, 100)
coordG = Vec2d(300, 100)
# coordinate at midpoint of points G, F, B, H initially
coordD = coordG + VERT
coordF = coordD - HORIZ
coordH = coordD + HORIZ
coordB = coordD + VERT
coordK = coordH + HORIZ

ptA = staticBall(space, "A", coordA)
ptB = staticBall(space, "B", coordB)
ptC = staticBall(space, "C", coordC)
sliderG = ball(space, "G", coordG)
ballF = ball(space, "F", coordF)
ballH = ball(space, "H", coordH)
ballK = ball(space, "K", coordK)

beamAC = beam(space, ptA, ptC)
beamFG = beam(space, ballF, sliderG)
beamFB = beam(space, ballF, ptB)
beamHG = beam(space, ballH, sliderG)
beamHB = beam(space, ballH, ptB)
beamFK = beam(space, ballF, ballK, color=MUSTARD)

sliderD = ball(space, "D", coordD, color=RED)

tbeam, _, vLeft, vRight = \
  tBeam(space, sliderG, 80, 450, name="1", color=LIGHTGREEN)
    # supply arm & body lengths; get back body, up, left, right

# attach t-beam's arms to groove along beamAC
groove(space, beamAC, vRight)
groove(space, beamAC, vLeft)

groove(space, beamFK, ballH)

# add groove and sliderD to t-beam and beam
groove(space, tbeam, sliderD, (0,0), (0,450))
groove(space, beamFK, sliderD)

sliderG.velocity = (500, 0)  # move right

pts = animate(screen, space, sliderD)

# -------------------
# fancier manipulation than linkUtils's plotPoints()


# collect multiple y-values for each x-coord
coordsMap = {}
for (x,y) in pts:
  coordsMap.setdefault(x, []).append(y)

# average y-values for each x, and sort according to x-values
vs = []
for key, vals in coordsMap.items():
  vs.append((key, round(sum(vals)/len(vals))))
vs = sorted(vs)
# print(*vs)

# curve fit to (x,y)'s
xs, ys = zip(*vs)
paraArgs, _ = curve_fit(parabolaEqu, xs, ys)
a, b, c = paraArgs
equStr = f"y = {a:.5f}*x**2 + {b:.5f}*x + {c:.5f}"
print("Fitted:", equStr)

# plot points and fitted curve
plt.axis('equal')
plt.scatter(xs, ys, s=2, label="linkage points")
yCurve = [ parabolaEqu(x, a, b, c) for x in xs]
plt.plot(xs, yCurve, '--', color='red', label=equStr)
plt.legend()
plt.title(title)
plt.show()
