
# plantigrade.py
# Andrew Davison, ad@coe.psu.ac.th, Feb. 2025
# https://en.wikipedia.org/wiki/Chebyshev_lambda_linkage
# https://en.tcheb.ru/plantigrade-machine/
'''
  Four Chebyshev lambda linkages are driven by a 
  rotating parallelogram. Each linkage makes a 'leg'
  move forward and back in immitation of a walking step.
'''

import math
import os
os.environ['PYGAME_HIDE_SUPPORT_PROMPT'] = "hide"
from pymunk.vec2d import Vec2d
from linkUtils import *


def legCheb(space, a, coord0, kind, color):
  '''
    Generate the Chebyshev lambda linkage for driving
    a particular leg 'kind' ("IR", "OL", "IL", "OR"),
    and return the balls needed to attach the linkage
    to its leg and to the central parallelogram
  '''
  h = math.sqrt(6)

  if kind == "IR":
    coordA = coord0 + Vec2d(0.5*a, 0)  
    coordC = coordA + Vec2d(-2*a, 0)
    coordB = coordA + Vec2d(a, 0) 
    coordD = coordB + Vec2d(-1.5*a, -2*a)
    coordE = coordC + Vec2d(0, -4*a)
    ballA, ballB, ballC, ballE = \
      cheb(space, a, coordA, coordB, coordC, coordD, coordE, color)
    barBalls = (ballC, ballA, ballB, ballE)

  elif kind == "OL":
    coordA = coord0 + Vec2d(0.5*a, 0)  
    coordC = coordA + Vec2d(-2*a, 0)
    coordB = coordA + Vec2d(-a, 0) 
    coordD = coordB + Vec2d(-0.5*a, -h*a)
    coordE = coordC + Vec2d(0, -2*h*a)
    ballA, ballB, ballC, ballE = \
      cheb(space, a, coordA, coordB, coordC, coordD, coordE, color)
    barBalls = (ballC, ballB, ballA, ballE)

  elif kind == "IL":
    coordA = coord0 + Vec2d(-0.5*a, 0)  
    coordC = coordA + Vec2d(2*a, 0)
    coordB = coordA + Vec2d(-a, 0) 
    coordD = coordB + Vec2d(1.5*a, -2*a)
    coordE = coordC + Vec2d(0, -4*a)
    ballA, ballB, ballC, ballE = \
      cheb(space, a, coordA, coordB, coordC, coordD, coordE, color)
    barBalls = (ballB, ballA, ballC, ballE)

  elif kind == "OR":
    coordA = coord0 + Vec2d(-0.5*a, 0)  
    coordC = coordA + Vec2d(2*a, 0)
    coordB = coordA + Vec2d(a, 0) 
    coordD = coordB + Vec2d(0.5*a, -h*a)
    coordE = coordC + Vec2d(0, -2*h*a)
    ballA, ballB, ballC, ballE = \
      cheb(space, a, coordA, coordB, coordC, coordD, coordE, color)
    barBalls = (ballA, ballB, ballC, ballE)

  return barBalls



def cheb(space, a, coordA, coordB, coordC, coordD, coordE, c):
  '''
    create the balls and beams for a Chebyshev Lambda linkage
    at the specified coordinates.
    For a single version, see chebLambda.py
  '''
  ballA = ball(space, "", coordA)
  ballB = ball(space, "", coordB)
  ballC = ball(space, "", coordC)  
  ballD = ball(space, "", coordD)  
  ballE = ball(space, "", coordE)
  beamAB = beam(space, ballA, ballB, color=c)
  beamCD = beam(space, ballC, ballD, color=c)
  beamBE = beam(space, ballB, ballE, color=c)
  pin(space, beamBE, ballD, (coordE-coordB)/2)
  return ballA, ballB, ballC, ballE
         # ends of the linkage


def legsPair(space, topLeft, topRight, a, color=LIGHTBLUE):
  ''' 
    Use multiple segments to form two l-shaped legs 
    linked by a horizontal bridge. 
    e.g. see the kinematic scheme picture at the end of  https://en.tcheb.ru/plantigrade-machine/
  '''
  body = pymunk.Body()
  body.position = topLeft.position
  # left and right legs in the pair
  lpole = pymunk.Segment(body, (0,0), (0, 6*a), 5)
  lfoot = pymunk.Segment(body, (0,6*a), (a,6*a), 5)
  rpole = pymunk.Segment(body, (6*a,0), (6*a, 6*a), 5)
  rfoot = pymunk.Segment(body, (6*a,6*a), (5*a,6*a), 5)
  bridge = pymunk.Segment(body, (0,a), (6*a,a), 5)
  for seg in [lpole, lfoot, rpole, rfoot, bridge]:
    seg.color = color + (0,)
    seg.mass = 1
    seg.elasticity = 0; seg.friction = 1
    seg.filter = SHP_FILT
  ljoint = pymunk.PinJoint(body, topLeft, (0, 0), (0, 0))
  ljoint.collide_bodies=False
  rjoint = pymunk.PinJoint(body, topRight, (6*a, 0), (0, 0))
  rjoint.collide_bodies=False
  space.add(lpole, lfoot, rpole, rfoot, bridge, body, ljoint, rjoint)
  return body


# -------------------------------------------

screen = scrInit("Chebyshev Plantigrade")
scrWidth, scrHeight = screen.get_size()
space = pymunk.Space()
coord0 = Vec2d(scrWidth/2, scrHeight/2)

a = 50
coordA = coord0 + Vec2d(-3*a, 0)
coordB = coord0 + Vec2d(-a, 0)
coordC = coord0 + Vec2d(a, 0)
coordD = coord0 + Vec2d(3*a, 0)

coordE = coordB + Vec2d(a, 0)
coordF = coordE + Vec2d(2*a, 0)
coordG = coordF + Vec2d(-2*a, 0)
coordH = coordG + Vec2d(-2*a, 0)

ptA = staticBall(space, "A", coordA)
ptB = staticBall(space, "B", coordB)  
ptC = staticBall(space, "C", coordC)  
ptD = staticBall(space, "D", coordD)  

beamAD = beam(space, ptA, ptD)  # central bar

# parallelogram
ballE = ball(space, "E", coordE)
ballF = ball(space, "F", coordF)  
ballG = ball(space, "G", coordG)  
ballH = ball(space, "H", coordH)  

beamEF = beam(space, ballE, ballF)
beamGF = beam(space, ballG, ballF)
beamHG = beam(space, ballH, ballG)
beamHE = beam(space, ballH, ballE)

pin(space, beamHE, ptB, (coordE-coordH)/2)
pin(space, beamGF, ptC, (coordF-coordG)/2)


# build four linkages for driving the "IR", "OL", "IL", "OR" legs;
# attach them to the parallelogram
chebOL = legCheb(space, a, 
                 coord0 + Vec2d(-1.5*a,0), "OL", BLUE)
pin(space, chebOL[0], ptA)
pin(space, chebOL[1], ballH)
pin(space, chebOL[2], ptB)

chebOR = legCheb(space, a, 
                 coord0 + Vec2d(1.5*a,0), "OR", RED)
pin(space, chebOR[0], ptC)
pin(space, chebOR[1], ballF)
pin(space, chebOR[2], ptD)

chebIR = legCheb(space, a, 
                 coord0 + Vec2d(-1.5*a,0), "IR", GREEN)
pin(space, chebIR[0], ptA)
pin(space, chebIR[1], ptB)
pin(space, chebIR[2], ballE)

chebIL = legCheb(space, a, 
                 coord0 + Vec2d(1.5*a,0), "IL", LIGHTGRAY)
pin(space, chebIL[0], ballG)
pin(space, chebIL[1], ptC)
pin(space, chebIL[2], ptD)


# attach legs pairs to their linkages
legsPair0 = legsPair(space, chebOL[3], chebOR[3], a)   
      # uses blue and red linkages
legsPair1 = legsPair(space, chebIR[3], chebIL[3], a, color=LIGHTGREEN)      # uses green and gray linkages


ballE.velocity = (0, 500)   # move down
ballF.velocity = (0, 500)
ballH.velocity = (0, -500)  # move up
ballG.velocity = (0, -500)

animate(screen, space, chebOL[3], force=5000)
